(* $Id: client.ml,v 1.26 2018/07/30 12:52:20 heinz Exp $
 *
 * GTF (Generic Test Framework) Client
 * Copyright (C) 2004  Heinz Deinhart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

open Batteries
open Gtf
open Mucho_question

let mk_tmpfile () =
  let rec work nr =
    let name = Printf.sprintf "/tmp/.gtf-client.%i.%i" (Unix.getpid ()) nr
    in
      if Sys.file_exists name then
	work (nr + 1)
      else
	name
  in
    work 0

(*
let string2image content =
  let tmpname = mk_tmpfile ()
  in let och = open_out tmpname
  in
    output_string och content;
    close_out och;
    let i = Image.load content []
    in
      Unix.unlink tmpname
*)

class display_controller nc lva idnr = object(_)
  method setup (test : test) =
    let window = GWindow.window ~title:"GTF"
		   ~position:`CENTER_ALWAYS ~width:1024 ~height:768 ()
    in let main_vbox = GPack.vbox ~packing:window#add ~spacing:4 ()
    in let scrolled_window =
	GBin.scrolled_window ~hpolicy:`AUTOMATIC
	  ~packing:(main_vbox#pack ~expand:true ~fill:true) ()
    in let vbox = GPack.vbox ~packing:scrolled_window#add_with_viewport
		    ~spacing:4 ~border_width:4 ()
    in let hbox = GPack.hbox ~packing:main_vbox#pack ~spacing:4 ()
    in let abgabe_button = GButton.button ~label:" ABGABE "
			     ~packing:hbox#pack ()
    in let _ =
	GMisc.label ~text:(Printf.sprintf "%s %s (%s)"
			     (test#student#firstname ())
			     (test#student#lastname ()) (test#student#idnr ()))
	  ~packing:hbox#pack ()
    in let status_text = GMisc.label ~text:"" ~packing:hbox#pack ()
    in let _ = GBin.alignment ~packing:(hbox#pack ~expand:true) ()
    in let zeit_anzeige = GMisc.label ~text:"ZEITANZEIGE" ~packing:hbox#pack ()
    in let time_start = Unix.gettimeofday ()
       and test_time_start = test#time
    in let intermediate_callback () =
	test#touch ();
	 nc#send_msg (Intermediate (lva, idnr, ((test#student#archive ()),
						(test#archive ()))));
	 true
    in let toid2 =
         GMain.Timeout.add ~ms:(1000 * 15) ~callback:intermediate_callback
    in let timeout_callback () =
	let now = Unix.gettimeofday ()
	in let delta = now -. time_start
	in
	  if (test_time_start -. delta) < 0.0 then
	    begin
	      test#set_time 0.0;
	      GMain.Timeout.remove toid2;
	      test#touch ();
	      nc#send_msg (Finish (lva, idnr, ((test#student#archive ()),
					       (test#archive ()))));
	      window#misc#hide ();
	      false
	    end
	  else
	    begin
	      let gamma = (int_of_float (test_time_start -. delta))
	      in
		zeit_anzeige#set_text
		  ("Verbleibende Zeit: "^
		   (string_of_int (gamma / 60))^":"^
		   (string_of_int (gamma mod 60)));
		test#set_time (test_time_start -. delta);
		true
	    end;
    in let toid1 = GMain.Timeout.add ~ms:1000 ~callback:timeout_callback
    in let do_abgabe () =
	GMain.Timeout.remove toid1;
	 GMain.Timeout.remove toid2;
	 nc#send_msg (Finish (lva, idnr, ((test#student#archive ()),
					  (test#archive ()))));
	 window#misc#hide ()
    in let update_status_text (qcount: int) (sdef,smod,sdon: int * int * int) =
	let str = Printf.sprintf "Sie haben %i Fragen, davon %i nicht, %i teils und %i ganz ausgefuellt" qcount sdef smod sdon
	in
	  status_text#set_text str
    in
      test#set_display_status update_status_text;
      ignore (abgabe_button#connect#clicked ~callback:do_abgabe);
      test#populate_vbox vbox;
      test#check_status ();
      window#show ()
end

(* FIXME: abgewoehnen dass der dreck bei escape abkackt *)
class chooser_dialog dirs nc = object(_)
  initializer
    let window = GWindow.dialog ~position:`CENTER_ALWAYS
		   ~border_width:2 ~title:"GTF Chooser" ()
    in let hbox_lva =
	GPack.hbox ~border_width:2 ~spacing:2 ~packing:window#vbox#pack ()
    in let hbox_idnr =
	GPack.hbox ~border_width:2 ~spacing:2 ~packing:window#vbox#pack ()
    in let hbox_name =
	GPack.hbox ~border_width:2 ~spacing:2 ~packing:window#vbox#pack ()
    in let hbox_buttons =
	GPack.hbox ~border_width:2 ~spacing:10
	  ~packing:window#action_area#add ()
    in let _ = GMisc.label ~text:"LVAs:" ~packing:hbox_lva#pack ()
    in let _ = GMisc.label ~text:"Matrikelnummer:" ~packing:hbox_idnr#pack ()
    in let _ = GMisc.label ~text:"Nachname:" ~packing:hbox_name#pack ()
    in let entry_lva = GEdit.combo ~popdown_strings:dirs ~allow_empty:false
			 ~packing:hbox_lva#pack ()
    in let entry_idnr = GEdit.entry ~text:"" ~packing:hbox_idnr#pack ()
    in let entry_name = GEdit.entry ~text:"" ~packing:hbox_name#pack ()
    in let login_callback = fun () ->
      nc#send_msg (RequestTest (entry_lva#entry#text, entry_idnr#text,
				entry_name#text));
      window#misc#hide ()
	(* todo: destroy window *)
    in let may_quit = fun () ->
      let md = GWindow.message_dialog
	~message:"You are about to quit the GTF application. This should only be done by supervisors at the end of the last exam slot. Do you really want to quit?"
	~message_type:`QUESTION ~buttons:GWindow.Buttons.yes_no
	~position:`CENTER ~modal:true ()
      in
	if (md#run ()) = `YES
	then
	  GMain.quit ()
	else
	  md#destroy ()
    in
    ignore ((GButton.button ~label:"Login" ~packing:hbox_buttons#add ())
	      #connect#clicked ~callback:login_callback);
    ignore ((GButton.button ~label:"Quit" ~packing:hbox_buttons#add ())
              #connect#clicked ~callback:may_quit);
    entry_lva#misc#grab_focus ();
    window#show ()
end

class connector_dialog = object(_)
  method setup (callback : string -> string -> unit) =
    let window = GWindow.dialog ~position:`CENTER_ALWAYS
		   ~border_width:2 ~title:"GTF Connector" ()
    in let hbox_entry =
	GPack.hbox ~border_width:2 ~spacing:2 ~packing:window#vbox#pack ()
    in let hbox_buttons =
	GPack.hbox ~border_width:2 ~spacing:10
	  ~packing:window#action_area#add ()
    in let _ = GMisc.label ~text:"Host:" ~packing:hbox_entry#pack ()
    in let entry_host = GEdit.entry ~text:"localhost"
			  ~packing:hbox_entry#pack ()
    in let _ = GMisc.label ~text:"Port:" ~packing:hbox_entry#pack ()
    in let entry_port =
	GEdit.entry ~max_length:5 ~width:50 ~text:"12018" ~packing:hbox_entry#pack ()
    in let entry_host_callback = fun () -> entry_port#misc#grab_focus ()
       and entry_port_callback =
	fun () ->
	  callback entry_host#text entry_port#text;
	  window#misc#hide ()
	  (* todo: destroy window *)
       in
    ignore (entry_host#connect#activate ~callback:entry_host_callback);
    entry_host#misc#grab_focus ();
    ignore (entry_port#connect#activate ~callback:entry_port_callback);
    ignore ((GButton.button ~label:"Connect" ~packing:hbox_buttons#add ())
              #connect#clicked ~callback:entry_port_callback);
    ignore ((GButton.button ~label:"Quit" ~packing:hbox_buttons#add ())
              #connect#clicked ~callback:GMain.quit);
    ignore (window#connect#destroy ~callback:GMain.quit);
    window#show ()
end

let question_generator csc = function
    "mucho_question" -> (new mucho_question csc false)#as_question ()
(*  | "word_question" -> (new word_question csc)#as_question () *)
  | unknown -> raise (Invalid_argument ("Question class: " ^ unknown))

class infotext_dialog (infotext, lva, idnr, test_archive) nc = object(_)
  initializer
  let start_test () =
	let test = new test question_generator
	in
	  test#unarchive test_archive;
	  (new display_controller nc lva idnr)#setup test;
	  ()
  in
  let show_infotext infotext =
    let dialog = GWindow.dialog ~position:`CENTER_ALWAYS ~modal:true
      ~border_width:2 ~title:"Information" ~show:true ()
    in let _ = GMisc.label ~text:infotext ~line_wrap:true
	~packing:dialog#vbox#pack ();
    in
      dialog#add_button "Abort" `ABORT;
      dialog#add_button "Start the test now" `START;
      if dialog#run () = `START then
	start_test ()
      else
	nc#send_msg (RequestDirectory);
      dialog#destroy ()
  in
    match infotext with
	Some x -> show_infotext x
      | None -> start_test ()
end

class network_client = object(self)
  inherit network_buffer

  val socket = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0

  method in_fd () =
    socket

  method out_fd () =
    socket

  method connect host_name port =
    try
      let host = Unix.gethostbyname host_name
      in let h_addr = host.Unix.h_addr_list.(0)
      in let sock_addr = Unix.ADDR_INET(h_addr,port)
      in
	Unix.connect socket sock_addr;
	dprintf 1 "connected to server"
    with
	_ ->
	  print_string (Printf.sprintf
			     "Failed to connect %s:%i\n" host_name port);
	  raise (Invalid_argument "")

  method setup host_name port =
    self#connect host_name port;
    ignore (GMain.Io.add_watch  ~callback:(self#handle_message) ~cond:[`IN]
              (GMain.Io.channel_of_descr socket));
    self#send_msg (RequestDirectory)

  method handle_message _ =
    dprintf 10 "handle_message ()";
    try
      begin
	match self#recv_msg () with
	    Directory dirs ->
	      ignore (new chooser_dialog dirs self)
	  | Test t ->
	      ignore (new infotext_dialog t self)
	  | ServerError str ->
	      dprintf 0 ("gtf: FATAL: handle_message: ServerError: "^str);
	      
	      let w = GWindow.dialog ~title:"ERROR" ~modal:true
			~position:`CENTER_ALWAYS ~border_width:2 ()
	      in
	      ignore ((GButton.button ~label:"Ack" ~packing:w#action_area#add ())
		        #connect#clicked ~callback:(fun () ->
		          self#send_msg
		            (RequestDirectory);
		          w#destroy ()));
	      ignore (GMisc.label ~text:str ~packing:w#vbox#pack ());
	      w#show ()
	  | _ ->
	      raise GTF_protocol_error
      end;
      true
    with
	End_of_file ->
	  GMain.Main.quit ();
	  true
end

class gtf_client = object(self)
  method connect host_name port_name =
    try
      (new network_client)#setup host_name (int_of_string port_name);
      ()
    with
	_ -> self#launch_connector ();

  method run ~host ~port ~autoconnect =
    ignore (GtkMain.Main.init ());
    begin
      match host, port, autoconnect with
      | (Some host), (Some port), true -> self#connect host port
      | _, _, _ -> self#launch_connector ()
    end;
    GMain.Main.main ()

  method launch_connector () =
    (new connector_dialog)#setup self#connect;
end

let parse_args () =
  let module P = BatOptParse in
  let p = P.OptParser.make ~prog:"gtfc" 
            ~usage:"%c - Generic Test Framework Client" () in
  let port_opt = P.StdOpt.str_option ()
    and server_opt = P.StdOpt.str_option ()
    and autocon_opt = P.StdOpt.store_true ()
  in
  P.OptParser.add p ~short_name:'s' ~long_name:"server"
    ~help:"hostname or ip of server" server_opt;
  P.OptParser.add p ~short_name:'p' ~long_name:"port"
    ~help:"port number of server" port_opt;
  P.OptParser.add p ~short_name:'a' ~long_name:"auto-connect"
    ~help:"automatically connect to server" autocon_opt;
  match P.OptParser.parse_argv p with
  | [] -> (new gtf_client)#run ~host:(P.Opt.opt server_opt)
            ~port:(P.Opt.opt port_opt)
            ~autoconnect:(P.Opt.get autocon_opt)
  | _ -> 
     Printf.fprintf stderr "%s: illegal extra arguments on command line\n"
       Sys.argv.(0);
     exit 1

let _ =
  parse_args ()

