(* $Id: gentest.ml,v 1.17 2006/04/23 20:42:56 heinz Exp $
 *
 * GTF (Generic Test Framework): test generator
 * Copyright (C) 2004  Heinz Deinhart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

(* msg, filename, linenr *)
exception File_format_error of string * string * int 

exception Fatal_error of string

exception Usage of string option

let die_usage_death msg =
  begin
    match msg with
	None -> ignore ()
      | Some msg -> Printf.fprintf stderr "%s: %s\n" Sys.argv.(0) msg
  end;
  Printf.fprintf stderr "%s: usage: %s " Sys.argv.(0) Sys.argv.(0);
  output_string stderr
    ("[-s] -t <time> -o <test-file> student-options question-list\n" ^
     "   where student-options are: -n <student-name>\n" ^
     "                              -m <student-number>\n" ^
     "   and question-list is: { -a <question-type> <question-args> } [..]\n" ^
     "   and following questions are currently implemented:\n" ^
     "     -a mucho_question <source> <question-nr> <answer-max>\n" ^
     "     -a word_question <source> <nr-of-questions>\n");
  exit 1

let question_generator csc = function
    "mucho_question" -> (new Mucho_question.mucho_question csc false)#as_question ()
(*  | "word_question" -> (new Word_question.word_question csc)#as_question () *)
  | unknown -> raise (Invalid_argument ("Question class: " ^ unknown))

(*
class word_db = object(self)
  val mutable questions : Word_question.word_question array = [||];

  method generate_questions count =
    Util.array_random_subarray questions count

  method read_from_file filename =
    let rec read_questions lnr chan q result stage =
      try
	let line = (input_line chan)
	in
	  match line.[0], line, stage with
	      '%', _, _ -> read_questions (lnr+1) chan q result stage
	    | '?', "?", 0 -> read_questions (lnr+1) chan q [] 1
	    | '?', "?", _ ->  read_questions (lnr+1) chan
		(new Word_question.word_question (fun () -> ()))
		  (q::result) 1
	    | '?', _, _ -> raise (File_format_error
				    ("word_question: illegal arg behind '?'",
				     filename, lnr))
	    | '=', _, _ ->
		q#set_agrep_errors (int_of_string (Str.string_after line 2));
		read_questions (lnr+1) chan q result 2
	    | _, newstr, _ ->
		match stage with
		    1 ->
		      q#set_question_text (q#question_text ^ " " ^ line);
		      read_questions (lnr+1) chan q result 1
		  | 2 ->
		      q#set_agrep_pattern line;
		      read_questions (lnr+1) chan q result 3
		  | 0 -> raise (File_format_error
				  ("text before first '?'", filename, lnr))
		  | _ -> raise (File_format_error
				  ("only one line after '=' allowed",
				   filename, lnr))
      with
	  Invalid_argument "String.sub" ->
	    raise (File_format_error ("illegal = argument", filename, lnr))
	| Invalid_argument _ -> read_questions (lnr+1) chan q result stage
	| End_of_file -> q::result
	| x -> raise x
    in
      try
	questions <- Array.of_list (read_questions 1 (open_in filename)
	  (new Word_question.word_question (fun () -> ())) [] 0)
      with
	  Sys_error msg ->
	    raise (Fatal_error ("could not open word_db ("^filename^"): "^msg))
end
 *)
type mucho_db_entry =
    Empty
  | Question of (string * bool)
  | Answer of (string * bool array)
  | Picture of (string * string option)

class mucho_db  = object(_)
  val mutable questions : Mucho_question.mucho_question array = [||];

  method generate_questions count answer_max =
    Array.iter (fun q -> q#limit_answers answer_max) questions;
    Util.array_random_subarray questions (if count = 0 then
					    Array.length questions
					  else
					    count)

  method generate_test (_ : int) =
    let test = new Gtf.test question_generator
    in
      Array.iter (fun q -> test#add_question (q#as_question ())) questions;
      test

  method read_from_file filename =
    let eval_ops = function "? radio" -> true | _ -> false
    and get_kanal str =
      try
	let namestart = String.index_from str 2 ' '
	in
	  ((String.sub str 2 (namestart - 2)),
	   (Some (Str.string_after str namestart)))
      with
	  Not_found -> ((Str.string_after str 2), None)
    and pmconv str =
      let pmtobool = function
	  '+' -> true
	| '-' -> false
	| _ -> raise (File_format_error ("illegal +/- line", filename, -1))
      in let rec work str i ar =
	  if (String.length str) = i then
	    ar
	  else
	    work str (i+1) (Array.append ar [|(pmtobool str.[i])|])
      in
	work str 0 [||]
    in let rec read_lines lnr chan entry entry_list =
	try
	  let line = (input_line chan)
	  in match entry, line.[0], line with
	      _, '%', _ -> read_lines (lnr+1) chan entry entry_list
	    | _, '#', str ->
		read_lines (lnr+1) chan entry
		  ((Picture (get_kanal str))::entry_list)
	    | Picture (_,_), _, _ -> raise (Invalid_argument "wrong #")
	    | Empty, '?', ops ->
		read_lines (lnr+1) chan
		  (Question ("", (eval_ops ops))) entry_list
	    | Empty, _, _ ->
		raise (File_format_error
			 ("trash when expecting '?'", filename, lnr))
	    | _, '?', ops -> read_lines (lnr+1) chan
		(Question ("", (eval_ops ops))) (entry::entry_list)
	    | _, '+', _
	    | _, '-', _ ->
		begin
		  try
		    read_lines (lnr+1) chan
		      (Answer ("", (pmconv line)))
		      (entry::entry_list)
		  with
		      File_format_error (a,b,_) ->
			raise (File_format_error (a,b,lnr))
		end
	    | Question (str, ops), _, _ ->
		read_lines (lnr+1) chan
		  (Question ((str ^ line ^ " "), ops)) entry_list
	    | Answer (str, sol), _, _ ->
		read_lines (lnr+1) chan (Answer ((str ^ line ^ " "), sol))
		  entry_list
	with
	    End_of_file ->
	      List.rev (entry::entry_list)
	  | Invalid_argument _ -> (* if line is "" *)
	      read_lines (lnr+1) chan entry entry_list
	  | x -> raise x
    in
      try
	let db = read_lines 1 (open_in filename) Empty []
	in let read_file fname =
	    try
	      let f = open_in fname
	      in let size = in_channel_length f
	      in let str = Bytes.create size
	      in
		really_input f str 0 size;
		str
	    with
		x -> Gtf.dprintf 0 "error reading file %s\n";
		  raise x
	in let rec gen_q_strs db q qs =
	    match db with
		Empty :: rest -> gen_q_strs rest q qs
	      | Question (text, ops) :: rest ->
		  let nq = new Mucho_question.mucho_question (fun () -> ()) ops
		  in
		    nq#set_question_text text;
		    gen_q_strs rest nq (q::qs)
	      | Answer (text, slist) :: rest ->
		  q#add_answer text slist;
		  gen_q_strs rest q qs
	      | Picture (filename, name) :: rest ->
		  q#add_picture (Bytes.to_string (read_file filename)) name;
		  gen_q_strs rest q qs
	      | [] -> q::qs
	in
	  questions <- Array.of_list
	    (match List.rev (gen_q_strs db
			       (new Mucho_question.mucho_question (fun () -> ()) false) []) with
		 _::real_list -> real_list | [] -> [])
      with
	  Invalid_argument _ ->
	    raise (File_format_error ("+/- multiple solution error",
				      filename, -1))
	| Sys_error msg ->
	    raise (Fatal_error ("could not open mucho_db ("^filename^
				"): "^msg))
end

class virtual test_generator = object(_)
  method make_test =
    ()
end
(*
let create_word_questions argqueue =
  let source = Queue.pop argqueue
  in let question_count = (int_of_string (Queue.pop argqueue))
  in let word_db = new word_db
  in
    try
      word_db#read_from_file source;
      word_db#generate_questions question_count
    with
	File_format_error (msg, fname, lnr) ->
	  Printf.fprintf stderr
	    "%s: file format error in '%s' at line %i: %s\n"
	    Sys.argv.(0) fname lnr msg;
	  raise (File_format_error (msg, fname, lnr))
      |	x ->
	  Printf.fprintf stderr "%s: creation of word_question failed.\n"
	    Sys.argv.(0);
	  raise x
 *)
let create_mucho_questions argqueue =
  let source = Queue.pop argqueue
  in let question_cont = (int_of_string (Queue.pop argqueue))
  in let answer_max = (int_of_string (Queue.pop argqueue))
  in let mucho_db = new mucho_db
  in
    try
      mucho_db#read_from_file source;
      mucho_db#generate_questions question_cont answer_max
    with
	x ->
	  Printf.fprintf stderr "%s: creation of mucho_question failed.\n"
	    Sys.argv.(0);
	  raise x

let create_questions argqueue =
  match Queue.pop argqueue with
      "mucho_question" ->
	Array.map (fun q -> q#as_question ()) (create_mucho_questions argqueue)
(*    | "word_question" ->
	Array.map (fun q -> q#as_question ()) (create_word_questions argqueue) *)
    | unknown ->
	raise (Usage (Some ("no such question: " ^ unknown)))
	
let rec parse_args argqueue (test: Gtf.test) filename =
  if not (Queue.is_empty argqueue) then
    try
      match Queue.pop argqueue with
	  "-o" ->
	    parse_args argqueue test (Some (Queue.pop argqueue))
	| "-n" ->
	    test#student#set_lastname (Queue.pop argqueue);
	    parse_args argqueue test filename
	| "-m" ->
	    test#student#set_idnr (Queue.pop argqueue);
	    parse_args argqueue test filename
	| "-s" ->
	    test#set_shuffle true;
	    parse_args argqueue test filename
	| "-a" ->
	    test#add_questions (Array.to_list (create_questions argqueue));
	    parse_args argqueue test filename
	| "-t" ->
	    test#set_time ((float_of_string (Queue.pop argqueue)) *. 60.0);
	    parse_args argqueue test filename
	| _ ->
	    raise Queue.Empty
    with
	Failure _ ->
	  raise (Usage (Some "-t: i need the time in minutes"))
      | Queue.Empty ->
	  raise (Usage None)
  else
    (test, filename)

let _ =
  Random.self_init ();
  let argqueue = Queue.create ()
  in
    Array.iter (fun s -> Queue.push s argqueue) Sys.argv;
    ignore (Queue.pop argqueue); (* discard argv[0] *)
    let test, filename =
      parse_args argqueue (new Gtf.test question_generator) None
    in
      try
	if filename = None then
	  raise (Usage (Some "missing -o (use -o - for stdout)"));
	let och = match filename with
	    Some "-" -> stdout
	  | Some str -> (open_out str)
	  | None -> raise (Usage None)
	in
	  test#shuffle_questions ();
	  Marshal.to_channel och
	    ((test#student#archive ()), (test#archive ())) [];
	  exit 0
      with
	  Usage msg ->
	    die_usage_death msg
	| Sys_error msg ->
	    match filename with
		None
	      | Some "-" ->
		  Printf.fprintf stderr "%s: error writing to stdout: %s\n"
		    Sys.argv.(0) msg;
	      | Some str ->
		  Printf.fprintf stderr "%s: error writing file %s: %s\n"
		    Sys.argv.(0) str msg;
