(* $Id: util.ml,v 1.2 2004/01/28 17:39:28 heinz Exp $
 *
 * GTF (Generic Test Framework): Generic Utilities
 * Copyright (C) 2004  Heinz Deinhart
 *
 * This code is released under the GNU General Public License version 2 or
 * later.
 *)

(* expects Random.self_init to be called
 *)
let array_randomize ar =
  let len = Array.length ar in
    for i = 0 to (len - 1) do
      let tmp = ar.(i)
      and rnd = Random.int len
      in
	ar.(i) <- ar.(rnd);
	ar.(rnd) <- tmp
    done

let array_randomize2 ar1 ar2 =
  let len = Array.length ar1
  in
    if len != (Array.length ar2) then
      raise (Invalid_argument "Util.array_randomize2");
    for i = 0 to (len - 1) do
      let tmp1 = ar1.(i)
      and tmp2 = ar2.(i)
      and rnd = Random.int len
      in
	ar1.(i) <- ar1.(rnd);
	ar2.(i) <- ar2.(rnd);
	ar1.(rnd) <- tmp1;
	ar2.(rnd) <- tmp2
    done

let array_randomize3 ar1 ar2 ar3 =
  let len = Array.length ar1
  in
    if len != (Array.length ar2) then
      raise (Invalid_argument "Util.array_randomize2");
    if len != (Array.length ar3) then
      raise (Invalid_argument "Util.array_randomize2");
    for i = 0 to (len - 1) do
      let tmp1 = ar1.(i)
      and tmp2 = ar2.(i)
      and tmp3 = ar3.(i)
      and rnd = Random.int len
      in
	ar1.(i) <- ar1.(rnd);
	ar2.(i) <- ar2.(rnd);
	ar3.(i) <- ar3.(rnd);
	ar1.(rnd) <- tmp1;
	ar2.(rnd) <- tmp2;
	ar3.(rnd) <- tmp3
    done

(* expects Random.self_init to be called
 * switches automatically between two different implementations (big/small)
 *)
let array_random_subarray ar count =
  let small input_count =
      if input_count < count then
	raise (Invalid_argument "array_random_subarray: element underflow");
      let rec work input input_size output =
	if input_size = (input_count - count) then
	  output
	else
	  let victim = Random.int input_size
	  in let tmp = input.(victim)
	  in
	    input.(victim) <- input.(input_size - 1);
	    work input (input_size - 1) (Array.append output [|tmp|]);
      in
	work (Array.copy ar) input_count [||]
  and big _ =
    let ar2 = Array.copy ar
    in
      array_randomize ar2;
      match (Array.length ar2) - count with
	  0 -> ar2
	| rest -> Array.sub ar2 (Random.int rest) count
  in let input_count = (Array.length ar)
  in
    if input_count = count || input_count < 100 || count > 1000 then
      big input_count
    else
      small input_count
